#include <stdio.h>
#include <string.h>
#pragma hdrstop

/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.11, 2004-03-29, <e2500@narod.ru>

   Supported camera (tested)
      E2500
      E4500
      E5000
      E995
      E885
      E775
      E990
      E950
      E900 - Thanks to Sakura Shan
      E2100
      E3100
      E4300
      C4040
      C4000
      C2020

   Main function

*/

#include <stdlib.h>
#include "tiff.h"
#include "file.h"
#include "raw2nef.h"

#define program "raw2nef v0.11, by paul69 <e2500@narod.ru> [compiled: " __DATE__ " " __TIME__ "]\n"

bool raw_to_nef(int n, char* a[]);

int main(int argc, char* argv[] )
{
   return raw_to_nef( argc, argv );
}

static void usage();
static char ofn_buf[_MAX_PATH];
static char efn_buf[_MAX_PATH];

static bool confirm_E2500(TCCDParam* ccd_pars, FILE* in)
{
   fseek( in, ccd_pars->file_size/2, 0 );
   unsigned size = 10*ccd_pars->row_length;
   unsigned test = 0;
   for(unsigned i=0; i<size; i += 3)
   {
      byte bitbuf[3];
      if( !fread( bitbuf, 3, 1, in) )
         return false;

      test |= (bitbuf[1]>>4)|(bitbuf[2]);
   }
   if( (test & 3) == 0 )
      return true;
   return false;
}

static
TCCDParam default_pars[] = {
   {1600+16, 1200+4, 0x2CE000, "YCGM", 2424, 12, 2500, TCCDParam::fInterlaced, confirm_E2500 }, // E2500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, 4500, TCCDParam::fInterlaced }, // E4500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, 4040, TCCDParam::fInterlaced }, // Camedia C4040
   {2560+16, 1920+4, 0x718000, "GMYC", 3864, 12, 5000, TCCDParam::fInterlaced }, // E5000
   {2048+16, 1536+4, 0x48D000, "MGCY", 3096, 12,  995, TCCDParam::fInterlaced }, // E995
   {2048+16, 1536+4, 0x48D000, "GMYC", 3096, 12,  885, TCCDParam::fInterlaced }, // E885
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10,  775, TCCDParam::fInterlaced }, // E775
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10, 2020, TCCDParam::fInterlaced }, // Camedia C2020
   {2048+16, 1536+4, 0x48D000, "GMYC", 3096, 12,  990, TCCDParam::fInterlaced }, // E990
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10,  950, TCCDParam::fInterlaced }, // E950
   {1280+16,  960+4, 0x182004, "CYMG", 1632, 10,  900, TCCDParam::fInterlaced }, // E900

   {2288, 1710, 0x599000, "BGGR", 3432, 12, 4300, TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // E4300 (BGGR)
   {2064, 1542, 0x48E000, "GBRG", 3096, 12, 3100, TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // E3100 (GBRG)
   {1616, 1206, 0x2CE000, "RGGB", 2424, 12, 2100, TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // E2100 (RGGB) ++
   {2304, 1718, 0x5AA000, "BGGR", 3456, 12, 4000, TCCDParam::fInterlaced|TCCDParam::fPrimaryColors }, // Olympus C4000 (BGGR) ++
//   {2260, 1700, 0x753FA0, "RGGB", 4520, 16, 4000, TCCDParam::fPrimaryColors }, // Casio QV-4000 (BGGR) KX821003.RAW
};

const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

bool Match( const char* FileName, const char* pattern )
{
   for(int i=0;; ++i)
   {
      if( !pattern[i] ) return true;
      
      if( pattern[i] == '#' )
      {
         if( FileName[i] >= '0' && FileName[i] <= '9' )
            continue;
         break;
      }
      if( pattern[i] >= 'A' && pattern[i] <= 'Z' )
      {
         if( (FileName[i] & ~0x20) == pattern[i] )
            continue;
         break;
      }
      
      if( FileName[i] == pattern[i] )
         continue;
      break;
   }
   return false;
}

char* CreateExifFileName(const char* FilePath, int NameOffset, int ExtOffset)
{
   if( NameOffset + 8 != ExtOffset )
      return NULL;

   const char* FileName = FilePath+NameOffset;
   const char* NameTemplate = 0;
   const char* FileNum = 0;

   if( Match( FileName, "DSCN####.") ) {
      NameTemplate = "DSCN%04d.JPG";
      FileNum = FileName+4; }
   else
   if( Match( FileName, "DSC#####.") ) {
      NameTemplate = "DSC%05d.JPG";
      FileNum = FileName+3; }
   else                       
   if( Match( FileName, "P#######.") ) {
      NameTemplate = "P%07d.JPG";
      FileNum = FileName+1; }

   if( !FileNum ) return NULL;
   if( !NameTemplate ) return NULL;

   int n = atoi( FileNum );
   if( n <= 0 ) return NULL;

   char* efn = efn_buf;
   memcpy(efn, FilePath, NameOffset);
   sprintf(efn + NameOffset, NameTemplate, n+1);
   return efn_buf;
}

bool ProcessFile(const char* ifn, char* ofn, char* efn, unsigned camera,
   bool verbose, bool copy_exif, bool endian, bool compatible )
{
   int i;

   if( verbose )
      printf("input: %s\n", ifn);

   if( !ofn || !efn )
   {
      const char* inpFileName = ifn;

      const char* ifn_slash = strrchr(inpFileName, '/');
      if( ifn_slash )
         inpFileName = ifn_slash+1;

      const char* ifn_bslash = strrchr(inpFileName, '\\');
      if( ifn_bslash )
         inpFileName = ifn_bslash+1;

      const char* inpFileExt = inpFileExt = strrchr(inpFileName, '.');

      if( !inpFileExt )
         inpFileExt = ifn + strlen(ifn);

      if( !ofn )
      {
         ofn = ofn_buf;
         int len = inpFileExt - ifn;
         memcpy(ofn, ifn, len);
         strcpy(ofn+len, ".NEF");
      }

      if( !efn && copy_exif )
      {
         efn = CreateExifFileName(ifn, inpFileName-ifn, inpFileExt-ifn);
         if( !efn && verbose )
            printf("warning: cannot get exif filename\n");
      }
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - cannot open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   NEF_IFD Nef_IFD;

   if( efn )
   {
      if( verbose )
         printf("read exif info from \"%s\"\n", efn);

      TFile exif = fopen( efn, "rb");

      if( !exif && verbose )
         { printf("warning: \"%s\" - cannot open exif file\n", efn); }
      else
      if( !Nef_IFD.read_exif( exif ) && verbose )
         { printf("warning: \"%s\" - cannot read exif file\n", efn); }
   }

   IFDir::Tag* Model = Nef_IFD.IFD0.get_tag( TIFF::Model );
   if( Model && ((const char*)Model->value)[0] == 'E' )
   {
      unsigned exif_camera = atoi( (const char*)Model->value+1 );
      if( exif_camera != 0 )
      {
         if( camera == 0 )
            camera = exif_camera;
         else
         if( camera != exif_camera )
         { printf("warning: exif model E%u and option model E%u is not same\n", exif_camera, camera); }
      }
   }

   for(i=0; i<n_default_pars; ++i)
   {
      if( camera == 0 )
      {
         if( file_size == default_pars[i].file_size )
         {
            if( !default_pars[i].confirm )
               break;

            if( default_pars[i].confirm(&default_pars[i], in) )
               break;
         }
      }
      else
      {
         if( camera == default_pars[i].camera )
            break;
      }
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      if( verbose )
         printf("camera model is E%u\n", default_pars[i].camera );

      ccd_pars = default_pars[i];
   }
   else
   {
      printf("error: unsupported format\n");
      return false;
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read raw data from \"%s\"\n", ifn);

   if( !read_raw( in, ccd, ccd_pars) )
   {
      printf("error: can not read raw data\n");
      return false;
   }

   in.close();

   if( verbose )
      printf("write picture to: \"%s\"\n", ofn);

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( !Nef_IFD.write_nef(out, endian, ccd, ccd_pars, compatible) )
      { printf("error: can not write NEF\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}

bool raw_to_nef(int n, char* a[])
{
   // endian: true for Intel, false for Motorolla
   bool endian = true;
   bool compatible = false;
   bool copy_exif = true;
   char* ifn = NULL;
   char* ofn = NULL;
   char* efn = NULL;
   bool verbose = false;
   bool batch_mode = true;
   unsigned camera = 0;

   if( n <= 1 )
   {
      usage();
      return false;
   }

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'v': verbose = (a[i][2] != '-'); break;
         case 'c': compatible = (a[i][2] != '-'); break;

         case 'e':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               efn = a[++i];
            else
               goto missing;
            break;

         case 'o':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               ofn = a[++i];
            else
               goto missing;
            break;

         case '?': usage(); return false;

         case 'a': camera = 0; break;

         default:
            if( a[i][1] >= '1' && a[i][1] <= '9' )
               camera = atoi( a[i]+1 );
            else
               printf("warning: unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("warning: invalid switch: %s\n", a[i]);
            break;

         missing:
            printf("warning: missing argument for %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
      }
   }

   if( verbose )
      printf( program );

   if( !ifn )
   {
      printf("error: input file is missing\n\n");
      return false;
   }

   if( batch_mode )
   {
      for(i=1; i<n; ++i)
      {
         if( a[i][0] == '-' )
         {
            switch( a[i][1] )
            {
            case 'o':
            case 'e':
               if( i+1 < n && a[i+1][0] != '-' )
                  ++i;
               break;
            }
         }
         else
         {
            if( a[i][0] )
            {
               if( !ProcessFile(a[i], NULL, NULL, camera, verbose, copy_exif, endian, compatible) )
                  printf("cannot process \"%s\"\n\n", a[i]);
            }
         }
      }
   }
   else
   {
      if( !ProcessFile(ifn, ofn, efn, camera, verbose, copy_exif, endian, compatible) )
         printf("cannot process \"%s\"\n", ifn);
   }
   return true;
}

void usage()
{
   printf("Coolpix RAW data to E5700/E8700 NEF file converter\n");
   printf( program );

   printf("\nargs: [options] input_file1 [input_file2...]\n");

   printf("\noptions: camera\n");
   printf("   -auto - autodetect (default)\n");
   printf("   -2100 - Coolpix 2100\t");   printf("   -2500 - Coolpix 2500\n");
   printf("   -3100 - Coolpix 3100\t");   printf("   -4000 - Camedia C4000\n");
   printf("   -4040 - Camedia C4040\t");  printf("   -4300 - Coolpix 4300\n");
   printf("   -4500 - Coolpix 4500\t");   printf("   -5000 - Coolpix 5000\n");
   printf("   -900 - Coolpix 900\t");     printf("   -950 - Coolpix 950\n");
   printf("   -775 - Coolpix 775\t");     printf("   -2020 - Camedia C2020\n");
   printf("   -995 - Coolpix 995\n");
   printf("   -885 - Coolpix 885 (auto is wrong for 885!)\n");
   printf("   -990 - Coolpix 990 (auto is wrong for 990!)\n");

   printf("\noptions: processing\n");
   printf("   -v - verbose\n");
   printf("   -c - compatible with E5700/E8700\n");
   printf("   -e <filename> - exif jpeg file name (disable batch mode)\n");
   printf("   -o <filename> - nef output file name (disable batch mode)\n");

   printf("\nSupported cameras: " );
   printf("Nikon Coolpix: E2100, E2500, E3100, E4300, E4500, E5000, \nE995, E885, E775, E990, E950, E900");
   printf(" / Olympus Camedia: C4000, C4040, C2020");
}

