#include <stdio.h>
#include <string.h>
#pragma hdrstop
/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.04, 2003-01-26, <e2500@narod.ru>

   TIFF: Image File Directory management
*/

#include "tiff.h"

int TIFF::sizeof_value[n_types] = {0,1,1,2,4,8,1,1,2,4,8,4};


IFDir::~IFDir()
{
   for(int i=0; i<tag_count_; ++i)
      { delete array_[i]; }
   delete [] array_;
}

IFDir::Tag::Tag(int _tag, int _type, int _count, const void* _value)
   : tag(_tag), type(_type), count(_count), value(NULL)
{
   int size = get_size();
   value = new char[size];
   
   if( _value )
      memcpy( value, _value, size );
   else
      memset( value, 0, size );
}

IFDir::Tag* IFDir::add(int tag, int type, int count, const void* value)
{
   if( type >= TIFF::n_types || count == 0 )
      return NULL;

   int i;
   for(i=0; i<tag_count_; ++i)
   {
      if( array_[i]->tag == tag )
      {
         delete array_[i];
         array_[i] = new Tag(tag, type, count, value);
         return array_[i];
      }
      if( array_[i]->tag > tag )
         break;
   }

   if( tag_count_ == arr_size_ )
   {
      int new_size = arr_size_ + 64;
      Tag** temp = array_;
      array_ = new Tag*[new_size];

      if( !array_ )
      {
         array_ = temp;
         return NULL;
      }

      arr_size_ = new_size;

      if( tag_count_ )
         memcpy( array_, temp, sizeof(array_[0])*tag_count_ );

      delete [] temp;
   }

   if( i < tag_count_ )
   {
      memmove( array_+i+1, array_+i, sizeof(array_[0])*(tag_count_-i) );
   }

   array_[i] = new Tag(tag, type, count, value);

   ++tag_count_;
   return array_[i];
}

unsigned IFDir::get_size()
{
   unsigned size = 2 + 12*tag_count_ + 4;
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];
      unsigned tag_data_size = t->get_size();
      if( tag_data_size > 4 )
         size += tag_data_size;
   }
   return size;
}

IFDir::Tag* IFDir::get_tag(word tag)
{
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];

      if( t->tag == tag )
         return t;
   }
   return NULL;
}

unsigned IFDir::get_tag_offset(word tag)
{
   unsigned offset = 2 + 12*tag_count_ + 4;
   for(int i=0; i<tag_count_; ++i)
   {
      Tag* t = array_[i];
      unsigned tag_data_size = t->get_size();

      if( t->tag == tag )
      {
         if( tag_data_size > 4 )
            return offset;

         return 2 + 12*i + 2+2+4; // tag+type+size
      }

      if( tag_data_size > 4 )
         offset += tag_data_size;
   }
   return 0;
}

